import os
import cocotb

from cocotb.triggers import (
    Timer,
    Edge,
)

from cocotb_test import simulator as cocotb_simulator

from cohdl import Entity, Port, BitVector, Signed, Unsigned
from cohdl import std

from ._proxy_port import ProxyPort


class Simulator:
    def __init__(
        self,
        entity: type[Entity],
        *,
        build_dir: str = "build",
        simulator: str = "ghdl",
        sim_args: list[str] | None = None,
        sim_dir: str = "sim",
        vhdl_dir: str = "vhdl",
        mkdir: bool = True,
        cast_vectors=None,
        extra_env: dict[str, str] | None = None,
        **kwargs,
    ):
        sim_args = [] if sim_args is None else sim_args
        extra_env = {} if extra_env is None else extra_env

        # This code is evaluated twice. Once in normal user code
        # to setup the test environment and again from another process
        # started by cocotb_simulator.run().
        # Use an environment variable to determine current mode.
        if os.getenv("COHDLSIM_TEST_RUNNING") is None:
            # running in normal user code
            # run CoHDL design into VHDL code and
            # start cocotb simulator

            sim_dir = f"{build_dir}/{sim_dir}"
            vhdl_dir = f"{build_dir}/{vhdl_dir}"

            if not os.path.exists(vhdl_dir):
                if mkdir:
                    os.makedirs(vhdl_dir, exist_ok=True)
                else:
                    raise AssertionError(
                        f"target directory '{vhdl_dir}' does not exist"
                    )

            lib = std.VhdlCompiler.to_vhdl_library(entity)

            top_name = lib.top_entity().name()
            entity_names = [sub.name() for sub in lib._entities]

            lib.write_dir(vhdl_dir)

            vhdl_sources = [f"{vhdl_dir}/{name}.vhd" for name in entity_names]

            # cocotb_simulator.run() requires the module name
            # of the Python file containing the test benches

            import inspect
            import pathlib

            filename = inspect.stack()[1].filename
            filename = pathlib.Path(filename).stem

            cocotb_simulator.run(
                simulator=simulator,
                sim_args=sim_args,
                sim_build=sim_dir,
                vhdl_sources=vhdl_sources,
                toplevel=top_name.lower(),
                module=filename,
                extra_env={"COHDLSIM_TEST_RUNNING": "True", **extra_env},
                **kwargs,
            )
        else:
            # running in simulator process
            # initialize members used by Simulator.test

            self._entity = entity
            self._dut = None
            self._port_bv = cast_vectors

    async def wait(self, duration: std.Duration):
        await Timer(int(duration.picoseconds()), units="ps")

    async def delta_step(self):
        await Timer(1, units="step")

    async def rising_edge(self, signal: ProxyPort):
        await signal._rising_edge()
        await self.delta_step()

    async def falling_edge(self, signal: ProxyPort):
        await signal._falling_edge()
        await self.delta_step()

    async def any_edge(self, signal: ProxyPort):
        await signal._edge()
        await self.delta_step()

    async def clock_cycles(self, signal: ProxyPort, num_cycles: int, rising=True):
        await signal._clock_cycles(num_cycles, rising)
        await self.delta_step()

    async def value_change(self, signal: ProxyPort):
        await Edge(signal._cocotb_port)

    async def value_true(self, signal: ProxyPort):
        while not signal:
            await signal._edge()

    async def value_false(self, signal: ProxyPort):
        while signal:
            await signal._edge()

    async def true_on_rising(self, clk: ProxyPort, cond, *, timeout: int | None = None):
        while True:
            await self.rising_edge(clk)
            if cond():
                return

            if timeout is not None:
                assert timeout != 0, "timeout while waiting for condition"
                timeout -= 1

    async def true_on_falling(
        self, clk: ProxyPort, cond, *, timeout: int | None = None
    ):
        while True:
            await self.falling_edge(clk)
            if cond():
                return

            if timeout is not None:
                assert timeout != 0, "timeout while waiting for condition"
                timeout -= 1

    async def start(self, coro):
        await cocotb.start(coro)

    def start_soon(self, coro):
        cocotb.start_soon(coro)

    def gen_clock(
        self, clk, period_or_frequency: std.Duration = None, /, start_state=False
    ):
        if isinstance(period_or_frequency, (std.Frequency, std.Duration)):
            period = period_or_frequency.period()

        half = int(period.picoseconds()) // 2

        async def thread():
            nonlocal clk
            while True:
                clk <<= start_state
                await Timer(half, units="ps")
                clk <<= not start_state
                await Timer(half, units="ps")

        self.start_soon(thread())

    def get_dut(self):
        assert (
            self._dut is not None
        ), "get_dut may only be called from a testbench function running in cocotb"
        return self._dut

    def test(self, testbench):
        @cocotb.test()
        async def helper(dut):
            self._dut = dut
            entity_name = self._entity.__name__

            class EntityProxy(self._entity):
                def __new__(cls):
                    return object.__new__(cls)

                def __init__(self):
                    pass

                def __str__(self):
                    return entity_name

                def __repr__(self):
                    return entity_name

            for name, port in self._entity._info.ports.items():

                if self._port_bv is not None:
                    port_type = type(Port.decay(port))

                    if issubclass(port_type, BitVector) and not (
                        issubclass(port_type, (Signed, Unsigned))
                    ):
                        if self._port_bv is Unsigned:
                            port = port.unsigned
                        elif self._port_bv is Signed:
                            port = port.signed
                        else:
                            raise AssertionError(
                                f"invalid default vector port type {self._port_bv}"
                            )

                setattr(EntityProxy, name, ProxyPort(port, getattr(dut, name)))

            # first delta step to load default values
            await Timer(1, units="step")
            # await self.delta_step()

            # use proxy entity instead of cocotb dut
            # in testbench function
            await testbench(EntityProxy())

        return helper

    def freeze(self, port: ProxyPort):
        port.freeze()

    def release(self, port: ProxyPort):
        port.release()



from cohdl_sim_ghdl_interface import ObjHandle

from cohdl import Signal, Bit, BitVector, Unsigned, Signed
from cohdl_sim._generic_proxy_port import _GenericProxyPort


class ProxyPort(_GenericProxyPort):
    def __init__(self, entity_port: Signal, ghdl_handle: ObjHandle, sim):
        super().__init__(entity_port)
        self._handle = ghdl_handle
        self._sim = sim

    def _load(self):
        val = self._handle.get_binstr()

        if issubclass(self._type, (Bit, BitVector)):
            self._val._assign(val.upper())
        else:
            raise AssertionError(f"type {type(self._type)} not supported")

    def _store(self):
        if isinstance(self._val, (Unsigned, Signed)):
            self._handle.put_binstr(str(self._val.bitvector))
        else:
            self._handle.put_binstr(str(self._val))



from cohdl import Entity, Signal, Bit
from cohdl import std

from typing import overload

class Simulator:
    def __init__(
        self,
        entity: type[Entity],
        *,
        build_dir: str = "build",
        simulator: str = "ghdl",
        sim_args: list[str] | None = None,
        sim_dir: str = "sim",
        vhdl_dir: str = "vhdl",
        mkdir: bool = True,
        cast_vectors=None,
        extra_env: dict[str, str] | None = None,
        **kwargs,
    ):
        """
        Simulator takes a CoHDL entity, generates VHDL code from it
        and starts a cocotb test session.

        build_dir, sim_dir, vhdl_dir and mkdir are used to customize the
        location of build and simulation output.

        The other arguments are forwarded to cocotb_test.simulator.run().
        """

    async def wait(self, duration: std.Duration):
        """
        wait for a given simulation duration
        """

    async def delta_step(self):
        """
        run simulation for a short time to update output ports
        """

    async def rising_edge(self, signal: Signal[Bit]):
        """
        wait for rising edge on signal
        """

    async def falling_edge(self, signal: Signal[Bit]):
        """
        wait for falling edge on signal
        """

    async def any_edge(self, signal: Signal[Bit]):
        """
        wait for changes of signal
        """

    async def clock_cycles(self, signal: Signal[Bit], num_cycles: int, rising=True):
        """
        wait for a number of clock cycles,
        rising determines whether rising or falling edges are counted
        """

    async def value_change(self, signal: Signal):
        """
        wait until signal changes
        """

    async def value_true(self, signal: Signal):
        """
        wait until signal becomes truthy
        """

    async def value_false(self, signal: Signal):
        """
        wait until signal becomes falsy
        """

    async def true_on_rising(
        self, clk: Signal[Bit], cond, *, timeout: int | None = None
    ):
        """
        Wait until cond is true after a rising edge of the clock signal.
        `cond` can be a port or a callable taking no arguments returning a boolean value.
        Raises an exception if the condition remains false for more than timeout rising edges.
        """

    async def true_on_falling(
        self, clk: Signal[Bit], cond, *, timeout: int | None = None
    ):
        """
        Wait until cond is true after a falling edge of the clock signal.
        `cond` can be a port or a callable taking no arguments returning a boolean value.
        Raises an exception if the condition remains false for more than timeout falling edges.
        """

    async def start(self, coro):
        """
        Run coro in parallel task.
        `coro` will start immediately (current task is suspended).
        """

    def start_soon(self, coro):
        """
        Run coro in parallel task.
        `coro` will start once the current task is suspended.
        """

    def gen_clock(
        self,
        clk: Signal[Bit],
        period_or_frequency: std.Duration | std.Frequency,
        /,
        start_state=False,
    ) -> None:
        """
        Start a parallel task that produces a clock signal
        with the specified period or frequency on `clk`.
        """

    def get_dut(self):
        """
        return the cocotb design under test object
        """

    def test(self, testbench):
        """
        decorator turns coroutines into test benches
        """
        return testbench

    def freeze(self, port: Signal):
        """
        freeze the signal in its current state
        """

    def release(self, port: Signal):
        """
        release a frozen signal
        """



from cohdl import Entity, Port, BitVector, Unsigned, Signed
from cohdl import std

from pathlib import Path
from functools import partial
from ._build_simulation import prepare_ghdl_simulation
from ._proxy_port import ProxyPort

import os

from cohdl_sim_ghdl_interface import GhdlInterface


class _Suspend:
    def __await__(self):
        yield None


_suspend = _Suspend()

global_alive_list = []


# ugly workaround to keep python objects referenced in the simulator alive
def _keep_alive(i):
    global_alive_list.append(i)
    return i


class Simulator:
    def __init__(
        self,
        entity: type[Entity],
        *,
        build_dir: str = "build",
        mkdir=True,
        sim_args=None,
        cast_vectors=None,
    ):
        build_dir = Path(build_dir)

        if not os.path.exists(build_dir):
            if mkdir:
                os.makedirs(build_dir, exist_ok=True)
            else:
                raise AssertionError(f"target directory '{build_dir}' does not exist")

        lib = std.VhdlCompiler.to_vhdl_library(entity)

        top_name = lib.top_entity().name()
        entity_names = [sub.name() for sub in lib._entities]

        lib.write_dir(build_dir)

        vhdl_sources = [f"{build_dir}/{name}.vhd" for name in entity_names]

        self._entity = entity
        self._simlib = prepare_ghdl_simulation(
            vhdl_sources, top_name, build_dir, copy_files=False
        )

        self._top_name = top_name
        self._sim = GhdlInterface()
        self._sim_args = sim_args if sim_args is not None else []

        self._tb = None
        self._current_coro = None
        self._port_bv = cast_vectors

    def port_names(self) -> list[str]:
        return [name for name in self._entity._info.ports]

    def _initial_fn(self):
        entity_name = self._entity.__name__

        class EntityProxy(self._entity):
            def __new__(cls):
                return object.__new__(cls)

            def __init__(self):
                pass

            def __str__(self):
                return entity_name

            def __repr__(self):
                return entity_name

        for name, port in self._entity._info.ports.items():
            if self._port_bv is not None:
                port_type = type(Port.decay(port))

                if issubclass(port_type, BitVector) and not (
                    issubclass(port_type, (Signed, Unsigned))
                ):
                    if self._port_bv is Unsigned:
                        port = port.unsigned
                    elif self._port_bv is Signed:
                        port = port.signed
                    else:
                        raise AssertionError(
                            f"invalid default vector port type {self._port_bv}"
                        )

            setattr(
                EntityProxy,
                name,
                ProxyPort(
                    port, self._sim.handle_by_name(f"{self._top_name}.{name}"), sim=self
                ),
            )

        self._continue(self._tb(EntityProxy()))

    def _startup_function(self):
        # self._initial_fn()
        # global_alive_list.append(self._sim.add_startup_function(self._initial_fn))
        global_alive_list.append(self._sim.add_callback_delay(self._initial_fn, 0))

    def _continue(self, coro, name=None):
        prev_coro = self._current_coro
        self._current_coro = coro

        try:
            coro.send(None)
        except StopIteration:
            pass

        self._current_coro = prev_coro

    def test(self, testbench, sim_args=None):
        async def tb_wrapper(entity):
            await testbench(entity)
            self._sim.finish_simulation()

        sim_args = sim_args if sim_args is not None else self._sim_args

        self._tb = tb_wrapper
        self._sim.cleanup()

        global_alive_list.append(self._sim.add_startup_function(self._startup_function))
        self._sim.start(str(self._simlib), sim_args)
        self._sim.stop()

    async def _wait_picoseconds(self, picos: int):
        with _keep_alive(
            self._sim.cb_delay(
                partial(self._continue, self._current_coro, name="picos"), picos
            )
        ):
            await _suspend

    async def wait(self, duration: std.Duration):
        await self._wait_picoseconds(int(duration.picoseconds()))

    async def delta_step(self):
        with _keep_alive(
            self._sim.cb_delay(partial(self._continue, self._current_coro), 1)
        ):
            await _suspend

    async def rising_edge(self, signal: ProxyPort):
        with self._sim.cb_value_change(
            signal._root._handle,
            partial(self._continue, self._current_coro, name="rising_edge"),
        ):
            prev_state = signal.copy()

            while True:
                await _suspend
                new_state = signal.copy()

                if (not prev_state) and new_state:
                    await self.delta_step()
                    return
                prev_state = new_state

    async def falling_edge(self, signal: ProxyPort):
        with self._sim.cb_value_change(
            signal._root._handle,
            partial(self._continue, self._current_coro, name="falling_edge"),
        ):
            prev_state = signal.copy()

            while True:
                await _suspend
                new_state = signal.copy()

                if prev_state and not new_state:
                    await self.delta_step()
                    return
                prev_state = new_state

    async def any_edge(self, signal: ProxyPort):
        with self._sim.cb_value_change(
            signal._root._handle,
            partial(self._continue, self._current_coro, name="any_edge"),
        ):
            prev_state = signal.copy()

            while True:
                await _suspend
                new_state = signal.copy()

                if prev_state != new_state:
                    await self.delta_step()
                    return
                prev_state = new_state

    async def clock_cycles(self, signal: ProxyPort, num_cycles: int, rising=True):
        with self._sim.add_callback_value_change(
            signal._root._handle, partial(self._continue, self._current_coro)
        ):
            prev_state = signal.copy()

            if rising:
                while True:
                    await _suspend
                    new_state = signal.copy()

                    if (not prev_state) and new_state:
                        num_cycles -= 1
                        if num_cycles == 0:
                            await self.delta_step()
                            return

                    prev_state = new_state
            else:
                while True:
                    await _suspend
                    new_state = signal.copy()

                    if prev_state and not new_state:
                        num_cycles -= 1
                        if num_cycles == 0:
                            await self.delta_step()
                            return

                    prev_state = new_state

    async def value_change(self, signal: ProxyPort):
        with self._sim.add_callback_value_change(
            signal._root._handle, partial(self._continue, self._current_coro)
        ):
            await _suspend
            return

    async def value_true(self, signal: ProxyPort):
        with self._sim.add_callback_value_change(
            signal._root._handle, partial(self._continue, self._current_coro)
        ):
            while not signal:
                await _suspend

    async def value_false(self, signal: ProxyPort):
        with self._sim.add_callback_value_change(
            signal._root._handle, partial(self._continue, self._current_coro)
        ):
            while signal:
                await _suspend

    async def true_on_rising(self, clk: ProxyPort, cond, *, timeout: int | None = None):
        while True:
            await self.rising_edge(clk)
            if cond():
                return

            if timeout is not None:
                assert timeout != 0, "timeout while waiting for condition"
                timeout -= 1

    async def true_on_falling(
        self, clk: ProxyPort, cond, *, timeout: int | None = None
    ):
        while True:
            await self.falling_edge(clk)
            if cond():
                return

            if timeout is not None:
                assert timeout != 0, "timeout while waiting for condition"
                timeout -= 1

    async def start(self, coro):
        self.start_soon(coro)
        await self.delta_step()

    def start_soon(self, coro):
        async def wrapper():
            await self.delta_step()
            self._continue(coro)

        self._continue(wrapper())

    def gen_clock(self, clk: ProxyPort, period: std.Duration, start_state=False):
        if isinstance(period, std.Frequency):
            period = period.period()

        half = int(period.picoseconds()) // 2

        async def thread():
            nonlocal clk

            while True:
                clk <<= start_state
                await self._wait_picoseconds(half)
                clk <<= not start_state
                await self._wait_picoseconds(half)

        self.start_soon(thread())



from cohdl import Bit, BitVector, Signed, Unsigned, Signal, Port, TypeQualifierBase

from cocotb.handle import Freeze, Release
from cocotb.triggers import RisingEdge, FallingEdge, Edge, ClockCycles


def to_cocotb(cohdl_val):
    if isinstance(cohdl_val, Bit):
        return bool(cohdl_val)
    return cohdl_val.unsigned.to_int()


decay = TypeQualifierBase.decay


def load_all(*args):
    for arg in args:
        if isinstance(arg, ProxyPort):
            arg._load()


_prev_property = None


class ProxyPort(TypeQualifierBase):
    def __init__(
        self,
        entity_port: Signal[BitVector],
        cocotb_port,
        is_root=True,
        uid: int | None = None,
    ):
        self._val = Port.decay(entity_port)
        self._type = type(self._val)
        self._root = is_root
        self._cocotb_port = cocotb_port
        self._uid = id(self) if uid is None else uid

    def decay(self):
        return Port.decay(self._val)

    def __call__(self):
        return self

    def _is_root(self):
        return self._root

    def _to_type(self, inp):
        return self._type(inp)

    def _load(self):
        val = self._cocotb_port.value

        if issubclass(self._type, (Bit, BitVector)):
            self._val._assign(val.binstr.upper())
        else:
            raise AssertionError(f"type {type(self._type)} not supported")

    def _store(self):
        if isinstance(self._val, Bit):
            self._cocotb_port.value = bool(self._val)
        else:
            self._cocotb_port.value = self._val.unsigned.to_int()

    def copy(self):
        self._load()
        return self._val.copy()

    @property
    def signed(self):
        global _prev_property
        assert issubclass(self._type, BitVector)

        if issubclass(self._type, Signed):
            _prev_property = self
            return self

        result = ProxyPort(self._val.signed, self._cocotb_port, False, self._uid)
        result._load = self._load
        result._store = self._store
        _prev_property = result
        return result

    @signed.setter
    def signed(self, value):
        assert (
            value._uid is self._uid
        ), "direct assignment to .signed property not allowed use '<<=' operator"

    @property
    def unsigned(self):
        global _prev_property
        assert issubclass(self._type, BitVector)

        if issubclass(self._type, Unsigned):
            _prev_property = self
            return self

        result = ProxyPort(self._val.unsigned, self._cocotb_port, False, self._uid)
        result._load = self._load
        result._store = self._store
        _prev_property = result
        return result

    @unsigned.setter
    def unsigned(self, value):
        assert (
            value._uid is self._uid
        ), "direct assignment to .unsigned property not allowed use '<<=' operator"

    @property
    def bitvector(self):
        global _prev_property
        assert issubclass(self._type, BitVector)

        if not issubclass(self._type, (Signed, Unsigned)):
            _prev_property = self
            return self

        result = ProxyPort(self._val.bitvector, self._cocotb_port, False, self._uid)
        result._load = self._load
        result._store = self._store
        _prev_property = result
        return result

    @bitvector.setter
    def bitvector(self, value):
        assert (
            value._uid is self._uid
        ), "direct assignment to .bitvector property not allowed use '<<=' operator"

    def __getitem__(self, arg):
        assert issubclass(self._type, BitVector)

        if isinstance(arg, slice):
            assert isinstance(arg.start, int)
            assert isinstance(arg.stop, int)
            assert arg.step is None
            result = ProxyPort(self._val[arg], self._cocotb_port, False)
        else:
            assert isinstance(arg, int)
            result = ProxyPort(self._val[arg], self._cocotb_port, False)

        # Replace load and store methods with version
        # of root object. They always update all bits of
        # a port even if they are accessed via a slice.
        result._load = self._load
        result._store = self._store

        return result

    def __setitem__(self, arg, value):
        pass

    def __ilshift__(self, src):
        if isinstance(src, ProxyPort):
            src = src._val

        self._val._assign(src)
        self._store()
        return self

    @property
    def next(self):
        raise AssertionError("reading from .next property not allowed")

    @next.setter
    def next(self, value):
        self <<= value

    @property
    def type(self):
        return self._type

    def __bool__(self):
        self._load()
        return self._val.__bool__()

    def __index__(self):
        self._load()
        return self._val.__index__()

    def __eq__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._val.__eq__(other)

    def __gt__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._val.__gt__(other)

    def __lt__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._val.__lt__(other)

    def __ge__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._val.__ge__(other)

    def __invert__(self):
        self._load()
        return self._val.__invert__()

    def __le__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._val.__le__(other)

    def __add__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._val.__add__(other)

    def __sub__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._val.__sub__(other)

    def __and__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._val.__and__(other)

    def __or__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._val.__or__(other)

    def __xor__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._val.__xor__(other)

    def __matmul__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._val.__matmul__(other)

    def __radd__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._val.__radd__(other)

    def __rsub__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._val.__rsub__(other)

    def __rand__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._val.__rand__(other)

    def __ror__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._val.__ror__(other)

    def __rxor__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._val.__rxor__(other)

    def __rmatmul__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._val.__rmatmul__(other)

    def __str__(self):
        self._load()
        return str(self._val)

    def __repr__(self):
        self._load()
        return repr(self._val)

    def freeze(self):
        assert (
            self._is_root()
        ), "freeze may only be used on port objects, not on sub-bits/slices"
        self._cocotb_port.value = Freeze()

    def release(self):
        assert (
            self._is_root()
        ), "release may only be used on port objects, not on sub-bits/slices"
        self._cocotb_port.value = Release()

    async def _rising_edge(self):
        if self._root:
            await RisingEdge(self._cocotb_port)
        else:
            self._load()
            prev = bool(self._val)

            while True:
                await Edge(self._cocotb_port)

                self._load()
                current = bool(self._val)

                if current and not prev:
                    return
                prev = current

    async def _falling_edge(self):
        if self._root:
            await FallingEdge(self._cocotb_port)
        else:
            self._load()
            prev = bool(self._val)

            while True:
                await Edge(self._cocotb_port)

                self._load()
                current = bool(self._val)

                if prev and not current:
                    return
                prev = current

    async def _edge(self):
        if self._root:
            await Edge(self._cocotb_port)
        else:
            self._load()
            prev = self._val.copy()

            while True:
                await Edge(self._cocotb_port)

                self._load()
                current = self._val.copy()

                if prev != current:
                    return
                prev = current

    async def _clock_cycles(self, num_cycles, rising=True):
        if self._root:
            await ClockCycles(self._cocotb_port, num_cycles, rising)
        else:
            if rising:
                for _ in num_cycles:
                    await self._rising_edge()
            else:
                for _ in num_cycles:
                    await self._falling_edge()

    def __await__(self):
        async def gen():
            while True:
                if self:
                    return
                await self._edge()

        return gen().__await__()

    def resize(self, *args, **kwargs):
        return decay(self).resize(*args, **kwargs)
